"""DContacts KProperty"""


from typing import TYPE_CHECKING, Optional
from uuid import UUID

from pydantic import Field

from dsms.knowledge.properties.base import KProperty, KPropertyItem

if TYPE_CHECKING:
    from typing import Callable


class ContactInfo(KPropertyItem):
    """Contact info"""

    name: str = Field(..., description="Name of the contact person")
    email: str = Field(..., description="EMail of the contact person")
    user_id: Optional[UUID] = Field(
        None, description="User ID of the contact person"
    )


class ContactsProperty(KProperty):
    """KProperty for contacts"""

    # OVERRIDE
    @property
    def k_property_item(cls) -> "Callable":
        return ContactInfo

    # OVERRIDE
    def _add(self, item: ContactInfo) -> ContactInfo:
        """Side effect when a ContactInfo is added to the KProperty"""
        return item

    # OVERRIDE
    def _update(self, item: ContactInfo) -> ContactInfo:
        """Side effect when a ContactInfo is updated at the KProperty"""
        return item

    # OVERRIDE
    def _get(self, item: ContactInfo) -> ContactInfo:
        """Side effect when getting the ContactInfo for a specfic kitem"""
        return item

    # OVERRIDE
    def _delete(self, item: ContactInfo) -> None:
        """Side effect when deleting the ContactInfo of a KItem"""



"""Linked KItems KProperty"""


from typing import TYPE_CHECKING, Optional
from uuid import UUID

from pydantic import ConfigDict, Field, model_serializer

from dsms.knowledge.properties.base import KProperty, KPropertyItem
from dsms.knowledge.utils import _get_kitem

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Union

    from dsms import KItem


class LinkedKItem(KPropertyItem):
    """Data model of a linked KItem"""

    # OVERRIDE
    id: Optional[UUID] = Field(
        None,
        description="ID of the KItem to be linked",
    )
    source_id: Optional[UUID] = Field(
        None, description="Source ID of the KItem"
    )

    # OVERRIDE
    model_config = ConfigDict(exclude={})

    # OVERRIDE
    def __setattr__(self, index: int, item: "Any") -> None:
        """Add KItem to updated buffer."""
        if self._kitem and self.source_id not in self.context.buffers.updated:
            self.context.buffers.updated.update({self.source_id: self._kitem})
        super().__setattr__(index, item)

    # OVERRIDE
    @property
    def kitem(cls) -> "KItem":
        """KItem related to the KPropertyItem"""
        if not cls.id:
            raise ValueError("KItem not defined yet for KProperty")
        return _get_kitem(cls.source_id)

    # OVERRIDE
    @model_serializer
    def serialize(self):
        """Serialize KPropertItem"""
        return {key: str(value) for key, value in self.__dict__.items()}


class LinkedKItemsProperty(KProperty):
    """KProperty for linked KItems"""

    # OVERRIDE
    @property
    def k_property_item(cls) -> "Callable":
        return LinkedKItem

    # OVERRIDE
    def _add(self, item: LinkedKItem) -> LinkedKItem:
        """Side effect when an LinkedKItem is added to the KProperty"""
        if self.kitem:
            item = self.k_property_item(id=item.id, source_id=self.kitem.id)
        else:
            item = self.k_property_item(id=item.id)
        return item

    # OVERRIDE
    def _update(self, item: LinkedKItem) -> LinkedKItem:
        """Side effect when an LinkedKItem is updated at the KProperty"""
        return item

    # OVERRIDE
    def _get(self, item: LinkedKItem) -> LinkedKItem:
        """Side effect when getting the LinkedKItem for a specfic kitem"""
        return _get_kitem(item.id)

    # OVERRIDE
    def _delete(self, item: LinkedKItem) -> None:
        """Side effect when deleting the LinkedKItem of a KItem"""

    # OVERRIDE
    def __setitem__(
        self, index: int, item: "Union[Dict, KPropertyItem]"
    ) -> None:
        """Add or Update KPropertyItem and add it to the updated-buffer."""

        self._mark_as_updated()
        item = self._check_k_property_item(item)
        if self.kitem:
            item.source_id = self.kitem.id
        try:
            if self[index] != item:
                item = self._update(item)
        except IndexError:
            item = self._add(item)
        super().super().__setitem__(index, item)  # pylint: disable=no-member

    # OVERRIDE
    def _check_and_add_item(self, item: "Union[Dict, Any]") -> KPropertyItem:
        item = self._check_k_property_item(item)
        if self.kitem:
            item.source_id = self.kitem.id
        return self._add(item)

    # OVERRIDE
    @property
    def kitem(cls) -> "KItem":
        """KItem context of the field"""
        return cls._kitem

    # OVERRIDE
    @kitem.setter
    def kitem(cls, value: "KItem") -> None:
        """KItem setter"""
        cls._kitem = value
        for item in cls:
            item.source_id = cls.kitem.id



"""DSMS connection module"""

from typing import TYPE_CHECKING, Any, Dict, List

from dsms.apps.utils import _get_available_apps
from dsms.core.configuration import Configuration
from dsms.core.context import Context
from dsms.core.utils import _ping_dsms
from dsms.knowledge.sparql_interface import SparqlInterface
from dsms.knowledge.utils import _search

from dsms.knowledge.utils import (  # isort:skip
    _commit,
    _get_kitem,
    _get_kitem_list,
    _get_remote_ktypes,
)

if TYPE_CHECKING:
    from enum import Enum
    from typing import Optional

    from dsms.apps import App
    from dsms.core.context import Buffers
    from dsms.knowledge.kitem import KItem
    from dsms.knowledge.ktype import KType


class DSMS:
    """General class for connecting and interfacting with DSMS."""

    _context = Context

    def __init__(self, config: Configuration = None, **kwargs) -> None:
        """Initialize the DSMS object."""

        self._config = None
        self._context.dsms = self

        if config is not None and not kwargs:
            self.config = config
        elif config is None:
            self.config = Configuration(**kwargs)
        else:
            raise ValueError(
                """`config`-keyword is defined among others.
                The `config`-keyword is reserved for passing a config-object directly.
                Please specify kwargs for to be passed to the `Configuration`-object _OR_
                an instance of this `Configuration`-object directly."""
            )

        self._sparql_interface = SparqlInterface(self)
        self._ktypes = _get_remote_ktypes()

    def __getitem__(self, key: str) -> "KItem":
        """Get KItem from remote DSMS instance."""
        return _get_kitem(key)

    def __delitem__(self, kitem) -> None:
        """Stage an KItem for the deletion.
        WARNING: Changes only will take place after executing the `commit`-method
        """

        from dsms.knowledge.kitem import (  # isort:skip
            KItem,
        )

        if not isinstance(kitem, KItem):
            raise TypeError(
                f"Object must be of type {KItem}, not {type(kitem)}. "
            )
        kitem.context.buffers.deleted.update({kitem.id: kitem})

    def commit(self) -> None:
        """Commit and empty the buffers of the KItems to the DSMS backend."""
        _commit(self.buffers)
        self.buffers.created = {}
        self.buffers.updated = {}
        self.buffers.deleted = {}

    def search(
        self,
        query: "Optional[str]" = None,
        ktypes: "Optional[List[KType]]" = [],
        annotations: "Optional[List[str]]" = [],
        limit: int = 10,
        allow_fuzzy: "Optional[bool]" = True,
    ) -> "List[KItem]":
        """Search for KItems in the remote backend."""
        return _search(query, ktypes, annotations, limit, allow_fuzzy)

    @property
    def sparql_interface(cls) -> SparqlInterface:
        """Sparql interface of the DSMS instance."""
        return cls._sparql_interface

    @property
    def ktypes(cls) -> "Enum":
        """ "Enum of the KTypes defined in the DSMS instance."""
        return cls._ktypes

    @property
    def config(cls) -> Configuration:
        """Property returning the DSMS Configuration"""
        return cls._config

    @config.setter
    def config(self, value) -> None:
        """Property setter returning the DSMS Configuration"""
        if not isinstance(value, Configuration):
            raise TypeError(
                f"""The passed config-kwarg with value `{value}`
                is not of type `{Configuration}`, but of type {type(value)}."""
            )
        self._config = value
        verify_connection(self)

    @property
    def headers(cls) -> Dict[str, Any]:
        """Request headers for authorization"""
        if cls.config.token:
            header = {
                "Authorization": f"{cls.config.token.get_secret_value()}"
            }
        else:
            header = {}
        return header

    @property
    def kitems(cls) -> "List[KItem]":
        """KItems instanciated and available in the remote backend.
        WARNING: This will download _all_ KItems in the backend owned
        by the current user and may resolve into long response times.
        The default timeout for requests is defined under the
        `request_timeout`-attribute in the `Configuration`-class."""
        return _get_kitem_list()

    @property
    def apps(cls) -> "List[App]":
        """Return available KItem apps in the DSMS"""
        return _get_available_apps()

    @property
    def buffers(cls) -> "Buffers":
        """Return buffers of the DSMS session"""
        return cls._context.buffers

    @property
    def context(cls) -> "Context":
        """Return DSMS context"""
        return cls._context

    @classmethod
    def __get_pydantic_core_schema__(cls):
        """Get validator of the DSMS-object."""
        yield verify_connection


def verify_connection(dsms: DSMS) -> None:
    """Check if DSMS is valid."""
    if not isinstance(dsms, DSMS):
        raise TypeError(
            f"""The passed object for the dsms-connection
                is not of type {DSMS}."""
        )
    try:
        response = _ping_dsms()
        if not response.ok:
            raise ConnectionError(
                f"""Host with `{dsms.config.host_url}`
                gave a response with status code `{response.status_code}`"""
            )
    except Exception as excep:
        raise ConnectionError(
            f"Invalid DSMS instance: `{dsms.config.host_url}`"
        ) from excep
    return dsms



"""Conftest for DSMS-SDK"""
import json
from typing import TYPE_CHECKING
from urllib.parse import urljoin

import pytest
import responses

if TYPE_CHECKING:
    from typing import Any, Dict


class MockDB:
    """Mock database"""

    kitems = {
        "78dfd0c7-0348-412c-8eb6-c91a4d340477": {
            "name": "foo123",
            "id": "78dfd0c7-0348-412c-8eb6-c91a4d340477",
        },
        "82c7a5a8-0983-4830-8fcf-5c8d167a928b": {
            "name": "foo456",
            "id": "82c7a5a8-0983-4830-8fcf-5c8d167a928b",
        },
        "a00af336-e714-499a-9dbf-298651939e62": {
            "name": "foo789",
            "id": "a00af336-e714-499a-9dbf-298651939e62",
        },
        "698acdc5-dd97-4217-906e-2c0b44248c17": {
            "name": "bar123",
            "id": "698acdc5-dd97-4217-906e-2c0b44248c17",
        },
    }

    hdf5 = {
        "698acdc5-dd97-4217-906e-2c0b44248c17": [
            {
                "column_id": 0,
                "name": "bar123column",
            },
        ]
    }


@pytest.fixture(scope="function")
def custom_address(request) -> str:
    try:
        url = getattr(request, "param")
    except Exception:
        url = "https://www.example.org/"
    return url


@pytest.fixture(scope="function")
def mock_responses(custom_address) -> "Dict[str, Any]":
    kitems = urljoin(custom_address, "api/knowledge/kitems")
    ping = urljoin(custom_address, "api/knowledge/docs")

    return {
        ping: [
            {"method": responses.GET, "returns": {"status": 200, "json": {}}}
        ],
        kitems: [
            {"method": responses.GET, "returns": {"status": 200, "json": {}}}
        ],
    }


@pytest.fixture(scope="function")
def mock_callbacks(custom_address) -> "Dict[str, Any]":
    ktypes = urljoin(custom_address, "api/knowledge-type/")

    def return_ktypes(request):
        ktypes = [
            {
                "name": "organization",
                "id": "organization",
                "data_schema": {
                    "title": "Organization",
                    "type": "object",
                    "properties": {},
                },
            },
        ]
        header = {"content_type": "application/json"}
        return (200, header, json.dumps(ktypes))

    def return_kitems(request):
        # Extract 'id' parameter from the URL
        url_parts = request.url.split("/")
        item_id = url_parts[-1]

        # Your logic to generate a dynamic response based on 'item_id'
        # This is just a placeholder; you should replace it with your actual logic
        if item_id not in MockDB.kitems:
            return 404, {}, "KItem does not exist"
        else:
            return 200, {}, json.dumps(MockDB.kitems[item_id])

    def return_hdf5(request):
        # Extract 'id' parameter from the URL
        url_parts = request.url.split("/")
        item_id = url_parts[-1]

        # Your logic to generate a dynamic response based on 'item_id'
        # This is just a placeholder; you should replace it with your actual logic
        if item_id not in MockDB.hdf5:
            return 404, {}, "KItem does not exist"
        else:
            return 200, {}, json.dumps(MockDB.hdf5[item_id])

    def _get_kitems() -> "Dict[str, Any]":
        return {
            urljoin(custom_address, f"api/knowledge/kitems/{uid}"): [
                {
                    "method": responses.GET,
                    "returns": {
                        "content_type": "application/json",
                        "callback": return_kitems,
                    },
                }
            ]
            for uid in MockDB.kitems
        }

    def _get_hdf5() -> "Dict[str, Any]":
        return {
            urljoin(custom_address, f"api/knowledge/data_api/{uid}"): [
                {
                    "method": responses.GET,
                    "returns": {
                        "content_type": "application/json",
                        "callback": return_hdf5,
                    },
                }
            ]
            for uid in MockDB.kitems
        }

    return {
        ktypes: [
            {
                "method": responses.GET,
                "returns": {
                    "content_type": "application/json",
                    "callback": return_ktypes,
                },
            }
        ],
        **_get_kitems(),
        **_get_hdf5(),
    }


@pytest.fixture(autouse=True, scope="function")
def register_mocks(mock_responses, mock_callbacks, custom_address) -> str:
    for url, endpoints in mock_responses.items():
        for response in endpoints:
            responses.add(
                response["method"],
                url,
                **response["returns"],
            )
    for url, endpoints in mock_callbacks.items():
        for response in endpoints:
            responses.add_callback(
                response["method"],
                url,
                **response["returns"],
            )
    return custom_address


@pytest.fixture(autouse=True, scope="function")
def reset_dsms_context():
    from dsms.core.context import Context

    Context.dsms = None


@pytest.fixture(scope="function")
def get_mock_kitem_ids():
    return list(MockDB.kitems.keys())



"""DSMS apps models"""

import urllib.parse

from pydantic import BaseModel, Field

from dsms.core.utils import _perform_request


class App(BaseModel):
    """KItem app list"""

    filename: str = Field(
        ..., description="File name of the notebook in the DSMS."
    )
    basename: str = Field(
        ..., description="Base name of the notebook in the DSMS."
    )
    folder: str = Field(
        ..., description="Directory of the notebook in the DSMS."
    )

    def get(self, as_html: bool = False) -> str:
        """Download the jupyter notebook"""
        safe_filename = urllib.parse.quote_plus(self.filename)
        response = _perform_request(
            f"knowledge/api/apps/{safe_filename}",
            "get",
            params={"as_html": as_html},
        )
        if not response.ok:
            message = f"Something went wrong downloading app `{self.filename}`: {response.text}"
            raise RuntimeError(message)
        return response.text



"""Core utils of the DSMS core"""
import re
from typing import Any
from urllib.parse import urljoin
from uuid import UUID

import requests
from requests import Response


def _kitem_id2uri(kitem_id: UUID) -> str:
    "Convert a kitem id in the DSMS to the full resolvable URI"
    from dsms import Context

    return urljoin(str(Context.dsms.config.host_url), str(kitem_id))


def _uri2kitem_idi(uri: str) -> str:
    "Extract the kitem id from an URI of the DSMS"
    from dsms import Context

    return uri.replace(f"{Context.dsms.config.host_url}/", "").split("/")[0]


def _ping_dsms():
    """General check if the remote DSMS instance is up and running"""
    return _perform_request("api/knowledge/docs", "get")


def _perform_request(route: str, method: str, **kwargs: Any) -> Response:
    """Perform a general request for a certain route and with a certain method.
    Kwargs are general arguments which can be passed to the `requests.request`-function.
    """
    from dsms import Context

    dsms = Context.dsms
    response = requests.request(
        method,
        url=urljoin(str(dsms.config.host_url), route),
        headers=dsms.headers,
        timeout=dsms.config.request_timeout,
        verify=dsms.config.ssl_verify,
        **kwargs,
    )
    response.encoding = Context.dsms.config.encoding
    return response


def _snake_to_camel(snake_str: str, first_upper=False) -> str:
    """Convert a snare-cases string to a camel-cased string.
    Optionally, the first letter can be lowered."""
    camel = "".join(x.title() for x in snake_str.split("_"))
    if first_upper:
        camel = camel[0].upper() + camel[1:]
    else:
        camel = camel[0].lower() + camel[1:]
    return camel


def _name_to_camel(input_string):
    """Remove special characters and make a CamelCased-str"""
    words = re.findall(r"\w+", input_string)
    camel_case_words = [word.title() for word in words]
    camel_case_string = "".join(camel_case_words)
    return camel_case_string



"""Custom properties of a KItem"""


from typing import TYPE_CHECKING, Any, Optional
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, model_validator

if TYPE_CHECKING:
    from typing import Set, Union

    from dsms import Context, KItem


class CustomProperties(BaseModel):
    """Model for the custom properties of the KItem"""

    id: Optional[UUID] = Field(None, description="ID of the KItem")
    content: Any = Field({}, description="Constent of the custom kitem")
    kitem: Optional[Any] = Field(
        None, description="KItem related to the CustomProperties", exclude=True
    )

    model_config = ConfigDict(
        extra="forbid", exclude={"kitem", "id"}, validate_assignment=True
    )

    def __str__(self) -> str:
        """Pretty print the custom properties"""
        fields = ", ".join(
            [
                f"{key}={value}"
                for key, value in self.__dict__.items()
                if key not in self.exclude
            ]
        )
        return f"{self.__class__.__name__}({fields})"

    def __repr__(self) -> str:
        """Pretty print the custom properties"""
        return str(self)

    def __hash__(self) -> int:
        return hash(str(self))

    def add(self, content: "Union[dict, Any]") -> None:
        """Add data for custom properties"""
        if isinstance(content, type(None)):
            self.content = {}
        elif self.data_schema and isinstance(content, dict):
            self.content = self.data_schema(content)
        elif not self.data_schema and isinstance(content, dict):
            self.content = content
        elif self.data_schema:
            if isinstance(content, self.data_schema):
                self.content = content
            else:
                raise TypeError(
                    f"""Item `{content}` must be of type {self.data_schema} or {dict},
                    not `{type(content)}`."""
                )
        else:
            raise TypeError(
                f"""Item `{content}` must be of type {self.data_schema} or {dict},
                not `{type(content)}`."""
            )
        self._mark_as_updated()

    def _mark_as_updated(self) -> None:
        if self.kitem and self.id not in self.context.buffers.updated:
            self.context.buffers.updated.update({self.id: self.kitem})

    def delete(self) -> None:
        """Delete data for custom properties"""
        self.content = None
        self._mark_as_updated()

    def update(self, content: "Union[dict, Any]") -> None:
        """Update data for custom properties"""
        self.content = None
        self.add(content)

    def get(self) -> Any:
        """Get data for custom properties"""
        return self.content

    @model_validator(mode="before")
    @classmethod
    def validate_kitem(cls, data: Any) -> "KItem":
        """Validate the custom properties with respect to the KType of the KItem"""
        kitem = data.get("kitem")
        content = data.get("content")
        if kitem:
            data["id"] = kitem.id
            if kitem.ktype.data_schema:
                data["content"] = kitem.ktype.data_schema(**content)
        return data

    @property
    def id(cls) -> Optional[UUID]:
        """Identifier of the KItem related to the CustomProperies"""
        if not cls.kitem:
            raise ValueError("KItem not defined yet.")
        return cls.kitem.id  # pylint: disable=E1101

    @property
    def data_schema(cls):
        """Data schema related to the ktype of the associated kitem."""
        if not cls.kitem:
            raise ValueError("KItem not defined yet.")
        return cls.kitem.ktype.data_schema  # pylint: disable=E1101

    @property
    def context(cls) -> "Context":
        """Getter for Context"""
        from dsms import (  # isort:skip
            Context,
        )

        return Context

    @property
    def exclude(cls) -> "Optional[Set[str]]":
        """Fields to be excluded from the JSON-schema"""
        return cls.model_config.get("exclude")



"""Affiliation KProperty"""

from typing import TYPE_CHECKING

from pydantic import Field

from dsms.knowledge.properties.base import KProperty, KPropertyItem

if TYPE_CHECKING:
    from typing import Callable


class Affiliation(KPropertyItem):
    """Affiliation of a KItem."""

    name: str = Field(..., description="Name of the affiliation")


class AffiliationsProperty(KProperty):
    """Affiliations property"""

    # OVERRIDE
    @property
    def k_property_item(cls) -> "Callable":
        return Affiliation

    # OVERRIDE
    def _add(self, item: Affiliation) -> Affiliation:
        """Side effect when an affiliation is added to the KProperty"""
        return item

    # OVERRIDE
    def _update(self, item: Affiliation) -> Affiliation:
        """Side effect when an affiliation is updated at the KProperty"""
        return item

    # OVERRIDE
    def _get(self, item: Affiliation) -> Affiliation:
        """Side effect when getting the affiliation for a specfic kitem"""
        return item

    # OVERRIDE
    def _delete(self, item: Affiliation) -> None:
        """Side effect when deleting the affiliation of a KItem"""



"""DSMS Subgraph interface"""

from typing import TYPE_CHECKING

from dsms.knowledge.sparql_interface.utils import (
    _create_subgraph,
    _delete_subgraph,
    _get_subgraph,
    _update_subgraph,
)

if TYPE_CHECKING:
    from rdflib import Graph

    from dsms import DSMS


class Subgraph:
    """Subgraph interface for DSMS"""

    def __init__(self, dsms):
        """Initalize the Sparql interface"""
        self._dsms: "DSMS" = dsms

    def update(self, graph: "Graph", repository: str = "knowledge") -> None:
        """Update a subgraph in the DSMS"""
        _update_subgraph(graph, self._dsms.config.encoding, repository)

    def create(self, graph: "Graph", repository: str = "knowledge") -> None:
        """Create a subgraph in the DSMS"""
        _create_subgraph(graph, self._dsms.config.encoding, repository)

    def delete(self, identifier: str, repository: str = "knowledge") -> None:
        """Delete a subgraph in the DSMS"""
        _delete_subgraph(identifier, repository)

    def get(
        self,
        identifier: str,
        repository: str = "knowledge",
        is_kitem_id: bool = False,
    ) -> "Graph":
        """Get a subgraph from the DSMS"""
        return _get_subgraph(identifier, repository, is_kitem_id)



"""HDF5 Properties of a KItem"""
from typing import TYPE_CHECKING

import pandas as pd
from pydantic import Field

from dsms.knowledge.properties.base import KProperty, KPropertyItem
from dsms.knowledge.utils import _get_hdf5_column

if TYPE_CHECKING:
    from typing import Any, Callable, List


class Column(KPropertyItem):
    """Column of an HDF5 data frame"""

    column_id: int = Field(..., description="Column ID in the data frame")

    name: str = Field(
        ..., description="Name of the column in the data series."
    )

    def get(self) -> "List[Any]":
        """Download the data for the column in a time series"""
        return _get_hdf5_column(self.id, self.column_id)


class HDF5Container(KProperty):
    """HDF5 container of a data frame related to a KItem"""

    # OVERRIDE
    @property
    def k_property_item(cls) -> "Callable":
        return Column

    # OVERRIDE
    def _add(self, item: Column) -> Column:
        """Side effect when an Column is added to the KProperty"""
        return item

    # OVERRIDE
    def _update(self, item: Column) -> Column:
        """Side effect when an Column is updated at the KProperty"""
        return item

    # OVERRIDE
    def _delete(self, item: Column) -> None:
        """Side effect when deleting the Column of a KItem"""

    # OVERRIDE
    def _get(self, item: Column) -> Column:
        """Side effect when getting the Column for a specfic kitem"""
        return item

    def to_df(self) -> pd.DataFrame:
        """Return hdf5 as pandas DataFrame"""
        data = {column.name: column.get() for column in self}
        return pd.DataFrame.from_dict(data)



"""Pytest for basic KItem connection properties"""
import pytest
import responses


@responses.activate
def test_kitem_basic(custom_address, get_mock_kitem_ids):
    """Test KItem properties"""

    from dsms.core.configuration import Configuration
    from dsms.core.context import Context
    from dsms.core.dsms import DSMS
    from dsms.knowledge.kitem import KItem

    assert Context.dsms is None

    with pytest.warns(UserWarning, match="No authentication details"):
        dsms = DSMS(host_url=custom_address)

    instance = KItem(
        id=get_mock_kitem_ids[0],
        name="foo123",
        ktype_id=dsms.ktypes.Organization,
    )

    assert isinstance(instance.dsms, DSMS)
    assert isinstance(instance.dsms.config, Configuration)
    assert Context.dsms == instance.dsms


@responses.activate
def test_kitem_config_class(custom_address, get_mock_kitem_ids):
    """Test KItem properties"""

    from dsms.core.configuration import Configuration
    from dsms.core.context import Context
    from dsms.core.dsms import DSMS
    from dsms.knowledge.kitem import KItem

    assert Context.dsms is None

    with pytest.warns(UserWarning, match="No authentication details"):
        config = Configuration(host_url=custom_address)
        dsms = DSMS(config=config)

    instance = KItem(
        id=get_mock_kitem_ids[0],
        name="foo123",
        ktype_id=dsms.ktypes.Organization,
    )

    assert isinstance(instance.dsms, DSMS)
    assert config == instance.dsms.config
    assert Context.dsms == instance.dsms


@responses.activate
def test_bad_config_kwargs(custom_address):
    from dsms import DSMS, Configuration

    with pytest.warns(UserWarning, match="No authentication details"):
        config = Configuration(host_url=custom_address)

    with pytest.raises(
        ValueError, match="`config`-keyword is defined among others."
    ):
        DSMS(host_url=custom_address, config=config)


@responses.activate
def test_kitem_custom_config_env(custom_address, get_mock_kitem_ids):
    import os

    from dsms.core.dsms import DSMS
    from dsms.knowledge.kitem import KItem

    os.environ["DSMS_HOST_URL"] = custom_address

    with pytest.warns(UserWarning, match="No authentication details"):
        dsms = DSMS()

    custom_instance = KItem(
        id=get_mock_kitem_ids[0],
        name="foo123",
        ktype_id=dsms.ktypes.Organization,
    )

    assert str(custom_instance.dsms.config.host_url) == custom_address

    os.environ.pop("DSMS_HOST_URL")


@responses.activate
def test_dsms_bad_object():
    from dsms.core.dsms import DSMS

    bad_object = "hello world"

    with pytest.raises(TypeError, match="The passed config-kwarg with value"):
        DSMS(config=bad_object)


@responses.activate
def test_kitem_bad_kwarg(get_mock_kitem_ids, custom_address):
    from pydantic import ValidationError

    from dsms.core.dsms import DSMS
    from dsms.knowledge.kitem import KItem

    with pytest.warns(UserWarning, match="No authentication details"):
        dsms = DSMS(host_url=custom_address)

    with pytest.raises(ValidationError):
        KItem(
            id=get_mock_kitem_ids[0],
            name="foo123",
            foo="bar",
            ktype_id=dsms.ktypes.Organization,
        )


@responses.activate
def test_kitem_connection_error():
    from dsms.core.configuration import Configuration
    from dsms.core.dsms import DSMS

    bad_url = "https://www.bad-dsms-instance.org"

    with pytest.warns(UserWarning, match="No authentication details"):
        bad_config = Configuration(host_url=bad_url)

    with pytest.raises(ConnectionError, match="Invalid DSMS instance:"):
        DSMS(config=bad_config)


@responses.activate
def test_kitem_default_ktypes(custom_address):
    from dsms.core.dsms import DSMS

    with pytest.warns(UserWarning, match="No authentication details"):
        dsms = DSMS(host_url=custom_address)

    assert len(dsms.ktypes) == 1


@responses.activate
def test_ktype_property(get_mock_kitem_ids, custom_address):
    from dsms.core.context import Context
    from dsms.core.dsms import DSMS
    from dsms.knowledge.kitem import KItem

    with pytest.warns(UserWarning, match="No authentication details"):
        dsms = DSMS(host_url=custom_address)

    kitem = KItem(
        id=get_mock_kitem_ids[0],
        name="foo123",
        ktype_id=dsms.ktypes.Organization,
    )

    assert kitem.ktype == Context.ktypes.get(dsms.ktypes.Organization.value)



"""Pytests for DSMS utilities"""

import pytest
import responses

# @responses.activate
# def test_kitem_list():
#     from dsms.core.dsms import DSMS

#     dsms = DSMS()


# @responses.activate
# def test_create_kitem():

#     from dsms.knowledge.utils import _create_new_kitem

#     from dsms.knowledge.kitem import KItem
#     from dsms.core.dsms import DSMS

#     dsms = DSMS()

#     item = KItem(name="foo", slug="bar", ktype_id=dsms.ktypes.Organization)


#     _create_new_kitem(item)


@responses.activate
def test_kitem_diffs(get_mock_kitem_ids, custom_address):
    from dsms.core.dsms import DSMS
    from dsms.knowledge.kitem import KItem
    from dsms.knowledge.utils import _get_kitems_diffs

    with pytest.warns(UserWarning, match="No authentication details"):
        dsms = DSMS(host_url=custom_address)

    linked_kitem1 = KItem(
        id=get_mock_kitem_ids[1],
        ktype_id=dsms.ktypes.Organization,
        name="foo456",
    )
    linked_kitem2 = KItem(
        id=get_mock_kitem_ids[2],
        ktype_id=dsms.ktypes.Organization,
        name="foo789",
    )
    linked_kitem3 = KItem(
        id=get_mock_kitem_ids[3],
        ktype_id=dsms.ktypes.Organization,
        name="bar123",
    )

    annotation = {
        "iri": "http://example.org/",
        "name": "foo",
        "namespace": "example",
    }
    annotation2 = {
        "iri": "http://example.org/",
        "name": "bar",
        "namespace": "example",
    }
    user_group = {"name": "private", "group_id": "private_123"}
    app = {"executable": "foo.exe"}
    app2 = {"executable": "bar.exe"}

    kitem_old = KItem(
        id=get_mock_kitem_ids[0],
        name="foo123",
        ktype_id=dsms.ktypes.Organization,
        annotations=[annotation2],
        linked_kitems=[linked_kitem1, linked_kitem2],
        user_groups=[user_group],
        kitem_apps=[app2],
    )

    kitem_new = KItem(
        id=get_mock_kitem_ids[0],
        name="foo123",
        ktype_id=dsms.ktypes.Organization,
        annotations=[annotation],
        linked_kitems=[linked_kitem3],
        user_groups=[user_group],
        kitem_apps=[app],
    )

    expected = {
        "kitems_to_link": [
            obj.model_dump() for obj in kitem_new.linked_kitems
        ],
        "annotations_to_link": [
            obj.model_dump() for obj in kitem_new.annotations
        ],
        "user_groups_to_add": [],
        "kitem_apps_to_update": [
            obj.model_dump() for obj in kitem_new.kitem_apps
        ],
        "kitems_to_unlink": [
            obj.model_dump() for obj in kitem_old.linked_kitems
        ],
        "annotations_to_unlink": [
            obj.model_dump() for obj in kitem_old.annotations
        ],
        "user_groups_to_remove": [],
        "kitem_apps_to_remove": [
            obj.model_dump() for obj in kitem_old.kitem_apps
        ],
    }
    diffs = _get_kitems_diffs(kitem_old, kitem_new)
    assert sorted(diffs) == sorted(expected)



"""Annotations KProperty"""

from typing import TYPE_CHECKING

from pydantic import Field

from dsms.knowledge.properties.base import KProperty, KPropertyItem

if TYPE_CHECKING:
    from typing import Callable


class Annotation(KPropertyItem):
    """KItem annotation model"""

    iri: str = Field(..., description="IRI of the annotation")
    name: str = Field(..., description="Name of the annotation")
    namespace: str = Field(..., description="Namespace of the annotation")


class AnnotationsProperty(KProperty):
    """KProperty for annotations"""

    # OVERRIDE
    @property
    def k_property_item(cls) -> "Callable":
        """Annotation data model"""
        return Annotation

    # OVERRIDE
    def _add(self, item: Annotation) -> Annotation:
        """Side effect when an Annotation is added to the KProperty"""
        return item

    # OVERRIDE
    def _update(self, item: Annotation) -> Annotation:
        """Side effect when an Annotation is updated at the KProperty"""
        return item

    # OVERRIDE
    def _delete(self, item: Annotation) -> None:
        """Side effect when deleting the Annotation of a KItem"""

    # OVERRIDE
    def _get(self, item: Annotation) -> Annotation:
        """Side effect when getting the Annotation for a specfic kitem"""
        return item



"""Attachment KProperty"""

from typing import TYPE_CHECKING

from pydantic import Field

from dsms.knowledge.properties.base import KProperty, KPropertyItem
from dsms.knowledge.utils import _get_attachment

if TYPE_CHECKING:
    from typing import Callable


class Attachment(KPropertyItem):
    """Attachment uploaded by a  certain user."""

    name: str = Field(..., description="File name of the attachment")

    def download(self) -> str:
        """Download attachment file"""
        return _get_attachment(self.id, self.name)


class AttachmentsProperty(KProperty):
    """KProperty for managing attachments."""

    # OVERRIDE
    @property
    def k_property_item(cls) -> "Callable":
        return Attachment

    # OVERRIDE
    def _add(self, item: Attachment) -> Attachment:
        """Side effect when an Attachment is added to the KProperty"""
        return item

    # OVERRIDE
    def _update(self, item: Attachment) -> Attachment:
        """Side effect when an Attachment is updated at the KProperty"""
        return item

    # OVERRIDE
    def _delete(self, item: Attachment) -> None:
        """Side effect when deleting the Attachment of a KItem"""

    # OVERRIDE
    def _get(self, item: Attachment) -> Attachment:
        """Side effect when getting the Attachment for a specfic kitem"""
        return item



"""UserGroup KProperty"""

from typing import TYPE_CHECKING

from pydantic import Field

from dsms.knowledge.properties.base import KProperty, KPropertyItem

if TYPE_CHECKING:
    from typing import Callable


class UserGroup(KPropertyItem):
    """Users groups related to a KItem."""

    name: str = Field(..., description="Name of the user group")
    group_id: str = Field(..., description="ID of the user group")


class UserGroupsProperty(KProperty):
    """KProperty for user_groups"""

    @property
    def k_property_item(cls) -> "Callable":
        """UserGroup data model"""
        return UserGroup

    # OVERRIDE
    def _add(self, item: UserGroup) -> UserGroup:
        """Side effect when an UserGroup is added to the KProperty"""
        return item

    # OVERRIDE
    def _update(self, item: UserGroup) -> UserGroup:
        """Side effect when an UserGroup is updated at the KProperty"""
        return item

    # OVERRIDE
    def _get(self, item: UserGroup) -> UserGroup:
        """Side effect when getting the UserGroup for a specfic kitem"""
        return item

    # OVERRIDE
    def _delete(self, item: UserGroup) -> None:
        """Side effect when deleting the UserGroup of a KItem"""



"""Sparql interface utilities for the DSMS"""
import io
from typing import TYPE_CHECKING

from rdflib.plugins.sparql.results.jsonresults import JSONResult

from dsms.core.utils import _kitem_id2uri, _perform_request

if TYPE_CHECKING:
    from typing import Any, Dict, Optional, TextIO, Union

    from rdflib import Graph


def _sparql_query(query: str, repository: str) -> "Dict[str, Any]":
    """Submit plain SPARQL-query to the DSMS instance."""
    response = _perform_request(
        "api/knowledge/sparql",
        "post",
        data={"query": query},
        params={"repository": repository},
    )
    if not response.ok:
        raise RuntimeError(f"Sparql was not successful: {response.text}")
    try:
        response = response.json()
    except Exception as excep:
        raise RuntimeError(
            f"""Something went wrong fetching injecting sparql-query:
            \n
            `{query}`"""
        ) from excep
    return response


def _sparql_update(
    file_or_pathlike: "Union[str, TextIO]",
    encoding: str,
    repository: str,
) -> None:
    """Submit plain SPARQL-query to the DSMS instance."""
    response = _perform_request(
        "api/knowledge/update-query",
        "post",
        files=_get_file_or_pathlike(file_or_pathlike, encoding),
        params={"repository": repository},
    )
    if not response.ok:
        raise RuntimeError(f"Sparql was not successful: {response.text}")


def _get_file_or_pathlike(
    file_or_pathlike: "Union[str, TextIO]", encoding: str
) -> "TextIO":
    if isinstance(file_or_pathlike, str):
        with open(file_or_pathlike, mode="r+", encoding=encoding) as file:
            files = {"file", file}
    else:
        if "read" not in dir(file_or_pathlike):
            raise TypeError(
                f"{file_or_pathlike} is neither a path"
                f"or a file-like object."
            )
        files = {"file": file_or_pathlike}
    return files


def _add_rdf(
    file_or_pathlike: "Union[str, TextIO]",
    encoding: str,
    repository: str,
    context: "Optional[str]" = None,
) -> None:
    """Create the subgraph in the remote backend"""
    params = {"repository": repository}
    if context:
        params["context"] = context
    response = _perform_request(
        "api/knowledge/add-rdf",
        "post",
        files=_get_file_or_pathlike(file_or_pathlike, encoding),
        params=params,
    )
    if not response.ok:
        raise RuntimeError(
            f"Not able to create subgraph in backend: {response.text}"
        )


def _delete_subgraph(identifier: str, repository: str) -> None:
    """Get subgraph related to a certain dataset id."""
    query = f"""
    DELETE {{
        ?s ?p ?o
    }}
    WHERE {{
        BIND(
            <{identifier}> as ?g
            )
        {{
            GRAPH ?g {{ ?s ?p ?o . }}
        }}
    }}"""
    response = _sparql_query(query, repository)
    if not response.get("boolean"):
        raise RuntimeError(
            f"Deleteing subgraph was not successful: {response}"
        )


def _create_subgraph(graph: "Graph", encoding: str, respository: str) -> None:
    """Create the subgraph in the remote backend"""
    upload_file = io.BytesIO(graph.serialize(encoding=encoding))
    _add_rdf(
        upload_file, encoding, respository, context=f"<{graph.identifier}>"
    )


def _update_subgraph(graph: "Graph", encoding: str, repository: str) -> None:
    """Update the subgraph in the remote backend"""
    _delete_subgraph(graph.identifier, repository)
    _create_subgraph(graph, encoding, repository)


def _get_subgraph(
    identifier: str, repository: str, is_kitem_id: bool = False
) -> "Graph":
    """Get subgraph related to a certain dataset id."""
    if is_kitem_id:
        identifier = _kitem_id2uri(identifier)
    query = f"""
    SELECT DISTINCT
        ?s ?p ?o
    WHERE {{
        BIND(
            <{identifier}> as ?g
            )
        {{
            GRAPH ?g {{ ?s ?p ?o . }}
        }}
    }}"""
    data = _sparql_query(query, repository)
    graph = JSONResult(data)
    graph.identifier = identifier

    if len(graph) == 0:
        raise ValueError(f"Subgraph for id `{identifier}` does not exist.")

    return graph



"""ExternalLink KProperty"""

from typing import TYPE_CHECKING

from pydantic import AnyUrl, Field

from dsms.knowledge.properties.base import KProperty, KPropertyItem

if TYPE_CHECKING:
    from typing import Callable


class ExternalLink(KPropertyItem):
    """External link of a KItem."""

    label: str = Field(..., description="Label of the external link")
    url: AnyUrl = Field(..., description="URL of the external link")


class ExternalLinksProperty(KProperty):
    """KProperty for external links"""

    # OVERRIDE
    @property
    def k_property_item(cls) -> "Callable":
        return ExternalLink

    # OVERRIDE
    def _add(self, item: ExternalLink) -> ExternalLink:
        """Side effect when an ExternalLink is added to the KProperty"""
        return item

    # OVERRIDE
    def _update(self, item: ExternalLink) -> ExternalLink:
        """Side effect when an ExternalLink is updated at the KProperty"""
        return item

    # OVERRIDE
    def _get(self, item: ExternalLink) -> ExternalLink:
        """Side effect when getting the ExternalLink for a specfic kitem"""
        return item

    # OVERRIDE
    def _delete(self, item: ExternalLink) -> None:
        """Side effect when deleting the ExternalLink of a KItem"""



"""Knowledge Item implementation of the DSMS"""

from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union
from urllib.parse import urljoin
from uuid import UUID, uuid4

import pandas as pd
from rdflib import Graph

from pydantic import (  # isort:skip
    BaseModel,
    ConfigDict,
    Field,
    ValidationInfo,
    field_validator,
)


from dsms.knowledge.properties import (  # isort:skip
    Affiliation,
    AffiliationsProperty,
    Annotation,
    AnnotationsProperty,
    App,
    AppsProperty,
    Attachment,
    AttachmentsProperty,
    Author,
    AuthorsProperty,
    ContactInfo,
    ContactsProperty,
    CustomProperties,
    ExternalLink,
    ExternalLinksProperty,
    KProperty,
    HDF5Container,
    Column,
    LinkedKItem,
    LinkedKItemsProperty,
    Summary,
    UserGroup,
    UserGroupsProperty,
)

from dsms.knowledge.ktype import KType  # isort:skip

from dsms.knowledge.utils import (  # isort:skip
    _kitem_exists,
    _slug_is_available,
    _slugify,
    _inspect_hdf5,
)

from dsms.knowledge.sparql_interface.utils import _get_subgraph  # isort:skip

if TYPE_CHECKING:
    from dsms import Context
    from dsms.core.dsms import DSMS


class KItem(BaseModel):
    """Knowledge Item of the DSMS."""

    # public
    name: str = Field(
        ..., description="Human readable name of the KContext.dsms"
    )
    id: Optional[UUID] = Field(
        default_factory=uuid4,
        description="ID of the KItem",
    )
    ktype_id: Union[Enum, str] = Field(..., description="Type ID of the KItem")
    slug: Optional[str] = Field(
        None, description="Slug of the KContext.dsms", min_length=4
    )
    annotations: List[Annotation] = Field(
        [], description="Annotations of the KItem"
    )
    attachments: List[Union[Attachment, str]] = Field(
        [],
        description="File attachements of the DSMS",
    )
    linked_kitems: List[Union[LinkedKItem, "KItem"]] = Field(
        [],
        description="KItems linked to the current KItem.",
    )
    affiliations: List[Affiliation] = Field(
        [],
        description="Affiliations related to a KItem.",
    )
    authors: List[Union[Author, str]] = Field(
        [], description="Authorship of the KItem."
    )
    avatar_exists: Optional[bool] = Field(
        False, description="Whether the KItem holds an avatar or not."
    )
    contacts: List[ContactInfo] = Field(
        [],
        description="Whether the KItem holds any contact information.",
    )
    created_at: Optional[Union[str, datetime]] = Field(
        None, description="Time and date when the KItem was created."
    )
    updated_at: Optional[Union[str, datetime]] = Field(
        None, description="Time and date when the KItem was updated."
    )
    external_links: List[ExternalLink] = Field(
        [],
        description="External links related to the KItem",
    )
    kitem_apps: List[App] = Field([], description="Apps related to the KItem.")
    summary: Optional[Union[str, Summary]] = Field(
        None, description="Human readable summary text of the KItem."
    )
    user_groups: List[UserGroup] = Field(
        [],
        description="User groups able to access the KItem.",
    )
    custom_properties: Optional[Any] = Field(
        {}, description="Custom properties associated to the KItem"
    )
    ktype: Optional[KType] = Field(
        None, description="KType of the KItem", exclude=True
    )

    hdf5: Optional[
        Union[List[Column], pd.DataFrame, Dict[str, Union[List, Dict]]]
    ] = Field(None, description="HDF5 interface.")

    model_config = ConfigDict(
        extra="forbid",
        validate_assignment=True,
        validate_default=True,
        exclude={"ktype"},
        arbitrary_types_allowed=True,
    )

    def __init__(self, **kwargs: "Any") -> None:
        """Initialize the KItem"""
        from dsms import DSMS

        # set dsms instance if not already done
        if not self.dsms:
            self.dsms = DSMS()

        # initalize the kitem
        super().__init__(**kwargs)

        # add kitem to buffer
        if (
            not _kitem_exists(self)
            and self.id not in self.context.buffers.created
        ):
            self.context.buffers.created.update({self.id: self})
            self.context.buffers.updated.update({self.id: self})

        self._set_kitem_for_properties()

    def __setattr__(self, name, value) -> None:
        """Add kitem to updated-buffer if an attribute is set"""
        super().__setattr__(name, value)
        self._set_kitem_for_properties()

        if (
            self.id not in self.context.buffers.updated
            and not name.startswith("_")
        ):
            self.context.buffers.updated.update({self.id: self})

    def __str__(self) -> str:
        """Pretty print the kitem fields"""
        fields = ", \n".join(
            [
                f"\n\t{key} = {value}"
                for key, value in self.__dict__.items()
                if key not in self.model_config["exclude"]
            ]
        )
        return f"{self.__class__.__name__}(\n{fields}\n)"

    def __repr__(self) -> str:
        """Pretty print the kitem Fields"""
        return str(self)

    def __hash__(self) -> int:
        return hash(str(self))

    @field_validator("affiliations")
    @classmethod
    def validate_affiliation(
        cls, value: List[Affiliation]
    ) -> AffiliationsProperty:
        """Validate affiliations Field"""
        return AffiliationsProperty(value)

    @field_validator("annotations")
    @classmethod
    def validate_annotations(
        cls, value: List[Annotation]
    ) -> AnnotationsProperty:
        """Validate annotations Field"""
        return AnnotationsProperty(value)

    @field_validator("attachments", mode="before")
    @classmethod
    def validate_attachments_before(
        cls, value: List[Union[str, Attachment]]
    ) -> List[Attachment]:
        """Validate attachments Field"""
        return [
            Attachment(name=attachment)
            if isinstance(attachment, str)
            else attachment
            for attachment in value
        ]

    @field_validator("attachments", mode="after")
    @classmethod
    def validate_attachments_after(
        cls, value: List[Attachment]
    ) -> AttachmentsProperty:
        """Validate attachments Field"""
        return AttachmentsProperty(value)

    @field_validator("kitem_apps")
    @classmethod
    def validate_apps(cls, value: List[App]) -> AppsProperty:
        """Validate apps Field"""
        return AppsProperty(value)

    @field_validator("authors")
    @classmethod
    def validate_authors(cls, value: List[Author]) -> AuthorsProperty:
        """Validate authors Field"""
        return AuthorsProperty(
            [
                Author(user_id=author) if isinstance(author, str) else author
                for author in value
            ]
        )

    @field_validator("contacts")
    @classmethod
    def validate_contacts(cls, value: List[ContactInfo]) -> ContactsProperty:
        """Validate contacts Field"""
        return ContactsProperty(value)

    @field_validator("external_links")
    @classmethod
    def validate_external_links(
        cls, value: List[ExternalLink]
    ) -> ExternalLinksProperty:
        """Validate external links Field"""
        return ExternalLinksProperty(value)

    @field_validator("linked_kitems", mode="before")
    @classmethod
    def validate_linked_kitems_list(
        cls, value: "List[Union[Dict, KItem, Any]]", info: ValidationInfo
    ) -> List[LinkedKItem]:
        """Validate each single kitem to be linked"""
        linked_kitems = []
        for item in value:
            if isinstance(item, dict):
                dest_id = item.get("id")
                if not dest_id:
                    raise ValueError("Linked KItem is missing `id`")
            elif isinstance(item, KItem):
                dest_id = item.id
            else:
                try:
                    dest_id = getattr(item, "id")
                except AttributeError as error:
                    raise AttributeError(
                        f"Linked KItem `{item}` has no attribute `id`."
                    ) from error
            linked_kitems.append(
                LinkedKItem(id=dest_id, source_id=info.data["id"])
            )
        return linked_kitems

    @field_validator("linked_kitems", mode="after")
    @classmethod
    def validate_linked_kitems(
        cls,
        value: List[LinkedKItem],
    ) -> LinkedKItemsProperty:
        """Validate the list out of linked KItems"""
        return LinkedKItemsProperty(value)

    @field_validator("user_groups")
    @classmethod
    def validate_user_groups(
        cls, value: List[UserGroup]
    ) -> UserGroupsProperty:
        """Validate user groups Field"""
        return UserGroupsProperty(value)

    @field_validator("custom_properties")
    @classmethod
    def validate_custom_properties(cls, value: Any) -> CustomProperties:
        """Validate custom properties Field"""
        if isinstance(value, dict) and "content" in value:
            value = CustomProperties(content=value["content"])
        elif isinstance(value, dict):
            value = CustomProperties(content=value)
        elif not isinstance(value, (CustomProperties, dict, type(None))):
            raise TypeError(
                f"""`custom_properties` must be of type {CustomProperties} or {dict},
                not {type(value)}."""
            )
        return value

    @field_validator("created_at")
    @classmethod
    def validate_created(cls, value: str) -> Any:
        """Convert the str for `created_at` in to a `datetime`-object"""
        from dsms import Context

        if isinstance(value, str):
            value = datetime.strptime(
                value, Context.dsms.config.datetime_format
            )
        return value

    @field_validator("updated_at")
    @classmethod
    def validate_updated(cls, value: str) -> Any:
        """Convert the str for `created_at` in to a `datetime`-object"""
        from dsms import Context

        if isinstance(value, str):
            value = datetime.strptime(
                value, Context.dsms.config.datetime_format
            )
        return value

    @field_validator("ktype")
    @classmethod
    def validate_ktype(cls, value: KType, info: ValidationInfo) -> KType:
        """Validate the data attribute of the KItem"""
        from dsms import Context

        ktype_id = info.data.get("ktype_id")

        if not value:
            if not isinstance(ktype_id, str):
                value = Context.ktypes.get(ktype_id.value)
            else:
                value = Context.ktypes.get(ktype_id)

            if not value:
                raise TypeError(
                    f"KType for `ktype_id={ktype_id}` does not exist."
                )

        return value

    @field_validator("slug")
    @classmethod
    def validate_slug(cls, value: str, info: ValidationInfo) -> str:
        """Validate slug"""
        ktype_id = info.data["ktype_id"]
        name = info.data.get("name")
        kitem_id = info.data.get("id")
        if not value:
            value = _slugify(name)
        slugified = _slugify(value)
        if len(slugified) < 4:
            raise ValueError("Slug length must have a minimum length of 4.")
        if value != slugified:
            raise ValueError(
                f"`{value}` is not a valid slug. A valid variation would be `{slugified}`"
            )
        if not _kitem_exists(kitem_id) and not _slug_is_available(
            ktype_id.value, value
        ):
            raise ValueError(f"Slug for `{value}` is already taken.")
        return value

    @field_validator("summary")
    @classmethod
    def validate_summary(cls, value: Union[str, Summary]) -> Summary:
        """Check whether the summary is a string or the dedicated model"""
        if isinstance(value, str):
            value = Summary(kitem=cls, text=value)
        return value

    @field_validator("hdf5")
    @classmethod
    def validate_hdf5(
        cls,
        value: Union[
            List[Column], pd.DataFrame, Dict[str, Dict[Any, Any]]
        ],  # pylint: disable=unused-argument
        info: ValidationInfo,
    ) -> HDF5Container:
        """Get HDF5 container if it exists."""
        kitem_id = info.data.get("id")
        if isinstance(value, (pd.DataFrame, dict)):
            if isinstance(value, pd.DataFrame):
                hdf5 = value.copy(deep=True)
            elif isinstance(value, dict):
                hdf5 = pd.DataFrame.from_dict(value)
            else:
                raise TypeError(
                    f"Data must be of type {dict} or {pd.DataFrame}, not {type(value)}"
                )
        else:
            columns = _inspect_hdf5(kitem_id)
            if columns:
                hdf5 = HDF5Container([Column(**column) for column in columns])
            else:
                hdf5 = None
        return hdf5

    def _set_kitem_for_properties(self) -> None:
        """Set kitem for CustomProperties and KProperties in order to
        remain the context for the buffer if any of these properties is changed.
        """
        for prop in self.__dict__.values():
            if (
                isinstance(prop, (KProperty, CustomProperties, Summary))
                and not prop.kitem
            ):
                prop.kitem = self

    @property
    def dsms(cls) -> "DSMS":
        """DSMS context getter"""
        return cls.context.dsms

    @dsms.setter
    def dsms(cls, value: "DSMS") -> None:
        """DSMS context setter"""
        cls.context.dsms = value

    @property
    def subgraph(cls) -> Optional[Graph]:
        """Getter for Subgraph"""
        return _get_subgraph(
            cls.id, cls.dsms.config.kitem_repo, is_kitem_id=True
        )

    @property
    def context(cls) -> "Context":
        """Getter for Context"""
        from dsms import (  # isort:skip
            Context,
        )

        return Context

    @property
    def url(cls) -> str:
        """URL of the KItem"""
        return urljoin(
            cls.context.dsms.config.host_url, f"{cls.ktype_id}/{cls.slug}"
        )



"""Sparql interface implementation for the DSMS"""

from typing import TYPE_CHECKING

from dsms.knowledge.sparql_interface.subgraph import Subgraph
from dsms.knowledge.sparql_interface.utils import (
    _add_rdf,
    _sparql_query,
    _sparql_update,
)

if TYPE_CHECKING:
    from typing import Any, Dict, TextIO, Union

    from dsms.core.dsms import DSMS


class SparqlInterface:

    """Sparql Interface for the DSMS."""

    def __init__(self, dsms):
        """Initalize the Sparql interface"""
        self._dsms: "DSMS" = dsms
        self._subgraph = Subgraph(dsms)

    def query(
        self, query: str, repository: str = "knowledge"
    ) -> "Dict[str, Any]":
        """Perform Sparql Query"""
        return _sparql_query(query, repository)

    def update(
        self,
        file_or_pathlike: "Union[str, TextIO]",
        repository: str = "knowledge",
    ) -> None:
        """Perform update query from local file"""
        _sparql_update(
            file_or_pathlike, self._dsms.config.encoding, repository
        )

    def insert(
        self,
        file_or_pathlike: "Union[str, TextIO]",
        repository: str = "knowledge",
    ) -> None:
        """Upload RDF to triplestore from local file"""
        _add_rdf(file_or_pathlike, self._dsms.config.encoding, repository)

    @property
    def subgraph(cls) -> Subgraph:
        """Subgraph interface for DSMS"""
        return cls._subgraph



"""DSMS knowledge utilities"""
import io
import json
import re
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union
from uuid import UUID

import pandas as pd
from pydantic import BaseModel, create_model
from requests import Response

from dsms.core.utils import _name_to_camel, _perform_request

if TYPE_CHECKING:
    from dsms.core.context import Buffers
    from dsms.knowledge import KItem, KType


def _parse_model(value: Dict[str, Any]) -> BaseModel:
    """Convert the dict with the model schema into a pydantic model."""
    title = value.get("title")
    properties = value.get("properties")
    if not title:
        raise KeyError("`data_schema` does not have any `title`.")
    if isinstance(properties, type(None)):
        raise KeyError("`data_schema` does not have any `properties`.")
    fields = {}
    for key, props in properties.items():
        dtype = props.get("type")
        default = props.get("default")
        if dtype == "string":
            dtype = str
        elif dtype == "integer":
            dtype = int
        elif dtype == "float":
            dtype = float
        elif dtype == "bool":
            dtype = bool
        elif dtype == "object":
            dtype = dict
        else:
            raise TypeError(f"Invalid `type={dtype}`")
        fields[key] = (dtype, ... if not default else default)
    return create_model(title, **fields)


def _get_remote_ktypes() -> Enum:
    """Get the KTypes from the remote backend"""
    from dsms import (  # isort:skip
        Context,
        KType,
    )

    response = _perform_request("api/knowledge-type/", "get")
    if not response.ok:
        raise ConnectionError(
            f"Something went wrong fetching the remote ktypes: {response.text}"
        )
    Context.ktypes = {ktype["id"]: KType(**ktype) for ktype in response.json()}

    return Enum("KTypes", {_name_to_camel(key): key for key in Context.ktypes})


def _get_kitem_list() -> "List[KItem]":
    """Get all available KItems from the remote backend."""
    from dsms.knowledge.kitem import (  # isort:skip
        KItem,
    )

    response = _perform_request("api/knowledge/kitems", "get")
    if not response.ok:
        raise ValueError(
            f"Something went wrong fetching the available kitems: {response.text}"
        )
    return [KItem(**kitem) for kitem in response.json()]


def _kitem_exists(kitem: Union[Any, str, UUID]) -> bool:
    """Check whether the KItem exists in the remote backend"""
    from dsms.knowledge.kitem import (  # isort:skip
        KItem,
    )

    if isinstance(kitem, KItem):
        route = f"api/knowledge/kitems/{kitem.id}"
    else:
        route = f"api/knowledge/kitems/{kitem}"
    response = _perform_request(route, "get")
    return response.ok


def _get_kitem(uuid: Union[str, UUID]) -> "KItem":
    """Get the KItem for a instance with a certain ID from remote backend"""
    from dsms import Context, KItem

    response = _perform_request(f"api/knowledge/kitems/{uuid}", "get")
    if response.status_code == 404:
        raise ValueError(
            f"""KItem with uuid `{uuid}` does not exist in
            DSMS-instance `{Context.dsms.config.host_url}`"""
        )
    if not response.ok:
        raise ValueError(
            f"""An error occured fetching the KItem with uuid `{uuid}`:
            `{response.text}`"""
        )
    return KItem(**response.json())


def _create_new_kitem(kitem: "KItem") -> None:
    """Create a new KItem in the remote backend"""
    payload = {
        "name": kitem.name,
        "id": str(kitem.id),
        "slug": kitem.slug,
        "ktype_id": kitem.ktype.id,
    }
    response = _perform_request("api/knowledge/kitems", "post", json=payload)
    if not response.ok:
        raise ValueError(
            f"KItem with uuid `{kitem.id}` could not be created in DSMS: {response.text}`"
        )


def _update_kitem(kitem: "KItem") -> Response:
    """Update a KItem in the remote backend."""
    old_kitem = _get_kitem(kitem.id)
    differences = _get_kitems_diffs(old_kitem, kitem)
    dumped = kitem.model_dump_json(
        exclude={
            "authors",
            "annotations",
            "linked_kitems",
            "updated_at",
            "avatar_exists",
            "user_groups",
            "ktype_id",
            "attachments",
            "id",
            "kitem_apps",
            "created_at",
            "external_links",
            "hdf5",
        },
        exclude_none=True,
    )
    payload = json.loads(dumped)
    payload.update(**differences)
    payload.update(
        external_links={
            link.label: str(link.url) for link in kitem.external_links
        }
    )
    response = _perform_request(
        f"api/knowledge/kitems/{kitem.id}", "put", json=payload
    )
    if not response.ok:
        raise ValueError(
            f"KItem with uuid `{kitem.id}` could not be updated in DSMS: {response.text}`"
        )
    for key, value in _get_kitem(kitem.id).__dict__.items():
        if key != "attachments":
            setattr(kitem, key, value)
    return response


def _delete_kitem(kitem: "KItem") -> None:
    """Delete a KItem in the remote backend"""
    response = _perform_request(f"api/knowledge/kitems/{kitem.id}", "delete")
    if not response.ok:
        raise ValueError(
            f"KItem with uuid `{kitem.id}` could not be deleted from DSMS: `{response.text}`"
        )


def _update_attachments(kitem: "KItem") -> None:
    """Update attachments of the KItem."""
    old_kitem = _get_kitem(kitem.id)
    differences = _get_attachment_diffs(old_kitem, kitem)
    for upload in differences["add"]:
        _upload_attachments(kitem, upload.name)
    for remove in differences["remove"]:
        _delete_attachments(kitem, remove.name)
    for key, value in _get_kitem(kitem.id).__dict__.items():
        setattr(kitem, key, value)


def _upload_attachments(kitem: "KItem", attachment: "str") -> None:
    """Upload the attachments of the KItem"""
    path = Path(attachment)

    if path.is_file():
        if not path.exists():
            raise FileNotFoundError(f"File {path} does not exist.")

        with open(path, mode="rb") as file:
            upload_file = {"dataFile": file}
            response = _perform_request(
                f"api/knowledge/attachments/{kitem.id}",
                "put",
                files=upload_file,
            )
        if not response.ok:
            raise RuntimeError(
                f"Could not upload attachment `{path}`: {response.text}"
            )


def _delete_attachments(kitem: "KItem", file_name: str) -> None:
    """Delete attachment from KItem"""
    url = f"api/knowledge/attachments/{kitem.id}/{file_name}"
    response = _perform_request(url, "delete")
    if not response.ok:
        raise RuntimeError(
            f"Could not delete attachment `{file_name}`: {response.text}"
        )


def _get_attachment(kitem_id: "KItem", file_name: str) -> str:
    """Download attachment from KItem"""
    url = f"api/knowledge/attachments/{kitem_id}/{file_name}"
    response = _perform_request(url, "get")
    if not response.ok:
        raise RuntimeError(
            f"Download for attachment `{file_name}` was not successful: {response.text}"
        )
    return response.text


def _get_attachment_diffs(kitem_old: "KItem", kitem_new: "KItem"):
    """Check which attachments should be removed and which should be added."""
    return {
        "remove": set(kitem_old.attachments) - set(kitem_new.attachments),
        "add": set(kitem_new.attachments) - set(kitem_old.attachments),
    }


def _get_kitems_diffs(kitem_old: "KItem", kitem_new: "KItem"):
    """Get the differences in the attributes between two kitems"""
    differences = {}
    attributes = [
        ("linked_kitems", ("kitems", "link", "unlink")),
        ("annotations", ("annotations", "link", "unlink")),
        ("kitem_apps", ("kitem_apps", "update", "remove")),
        ("user_groups", ("user_groups", "add", "remove")),
    ]
    for name, terms in attributes:
        to_add_name = terms[0] + "_to_" + terms[1]
        to_remove_name = terms[0] + "_to_" + terms[2]
        old_attr = getattr(kitem_old, name)
        new_attr = getattr(kitem_new, name)
        differences[to_add_name] = [
            json.loads(attr.model_dump_json())
            for attr in set(new_attr) - set(old_attr)
        ]
        differences[to_remove_name] = [
            json.loads(attr.model_dump_json())
            for attr in set(old_attr) - set(new_attr)
        ]
    return differences


def _commit(buffers: "Buffers") -> None:
    """Commit the buffers for the
    created, updated and deleted buffers"""
    _commit_created(buffers.created)
    _commit_updated(buffers.updated)
    _commit_deleted(buffers.deleted)


def _commit_created(buffer: "Dict[str, KItem]") -> dict:
    """Commit the buffer for the `created` buffers"""
    for kitem in buffer.values():
        exists = _kitem_exists(kitem)
        if not exists:
            _create_new_kitem(kitem)


def _commit_updated(buffer: "Dict[str, KItem]") -> None:
    """Commit the buffer for the `updated` buffers"""
    for kitem in buffer.values():
        if _kitem_exists(kitem):
            if isinstance(kitem.hdf5, pd.DataFrame):
                _update_hdf5(kitem.id, kitem.hdf5)
            elif isinstance(kitem.hdf5, type(None)) and _inspect_hdf5(
                kitem.id
            ):
                _delete_hdf5(kitem.id)
            _update_kitem(kitem)
            _update_attachments(kitem)


def _commit_deleted(buffer: "Dict[str, KItem]") -> None:
    """Commit the buffer for the `deleted` buffers"""
    for kitem in buffer.values():
        if _kitem_exists(kitem):
            _delete_hdf5(kitem.id)
            _delete_kitem(kitem)


def _search(
    query: Optional[str] = None,
    ktypes: "Optional[List[KType]]" = [],
    annotations: "Optional[List[str]]" = [],
    limit: "Optional[int]" = 10,
    allow_fuzzy: "Optional[bool]" = True,
) -> "List[KItem]":
    """Search for KItems in the remote backend"""
    from dsms import KItem  # isort:skip

    payload = {
        "search_term": query or "",
        "ktypes": [ktype.value for ktype in ktypes],
        "kitem_annotations": annotations,
        "limit": limit,
    }
    response = _perform_request(
        "api/knowledge/kitems/search",
        "post",
        json=payload,
        params={"allow_fuzzy": allow_fuzzy},
    )
    if not response.ok:
        raise RuntimeError(
            f"Search could not be executed successfully: {response.text}`"
        )
    try:
        dumped = response.json()
    except Exception as excep:
        raise RuntimeError(
            f"""Something went wrong while searching for KItems: {response.text}"""
        ) from excep
    return [KItem(**item) for item in dumped]


def _slugify(input_string):
    """Turn any arbitrary string into a slug."""
    slug = re.sub(
        r"[^\w\s]", "", input_string
    )  # Remove all non-word characters (everything except numbers and letters)
    slug = re.sub(r"\s+", "", slug)  # Replace all runs of whitespace
    slug = slug.lower()  # Convert the string to lowercase.
    return slug


def _slug_is_available(ktype_id: Union[str, UUID], value: str) -> bool:
    """Check whether the id of a KItem is available in the DSMS or not"""
    response = _perform_request(
        f"api/knowledge/kitems/{ktype_id}/{value}", "head"
    )
    return response.status_code == 404


def _get_hdf5_column(kitem_id: str, column_id: int) -> List[Any]:
    """Download the column of a hdf5 container of a certain kitem"""
    response = _perform_request(
        f"api/knowledge/data_api/{kitem_id}/column-{column_id}", "get"
    )
    if not response.ok:
        message = f"""Something went wrong fetch column id `{column_id}`
        for kitem `{kitem_id}`: {response.text}"""
        raise ValueError(message)
    return response.json().get("array")


def _inspect_hdf5(kitem_id: str) -> Optional[List[Dict[str, Any]]]:
    """Get column info for the hdf5 container of a certain kitem"""
    response = _perform_request(f"api/knowledge/data_api/{kitem_id}", "get")
    if not response.ok and response.status_code == 404:
        hdf5 = None
    elif not response.ok and response.status_code != 404:
        message = f"""Something went wrong fetching intospection
        for kitem `{kitem_id}`: {response.text}"""
        raise ValueError(message)
    else:
        hdf5 = response.json()
    return hdf5


def _update_hdf5(kitem_id: str, data: pd.DataFrame):
    buffer = io.BytesIO()
    data.to_json(buffer, indent=2)
    buffer.seek(0)
    response = _perform_request(
        f"api/knowledge/data_api/{kitem_id}", "put", files={"data": buffer}
    )
    if not response.ok:
        raise RuntimeError(
            f"Could not put dataframe into kitem with id `{kitem_id}`: {response.text}"
        )


def _delete_hdf5(kitem_id: str) -> Response:
    return _perform_request(f"api/knowledge/data_api/{kitem_id}", "delete")



"""General config for the DSMS Python SDK"""

import urllib
import warnings
from typing import Optional

import requests
from pydantic import AnyUrl, Field, SecretStr, field_validator
from pydantic_core.core_schema import ValidationInfo
from pydantic_settings import BaseSettings, SettingsConfigDict


class Configuration(BaseSettings):
    """General config for DSMS-SDK"""

    host_url: AnyUrl = Field(
        ..., description="Url of the DSMS instance to connect."
    )
    request_timeout: int = Field(
        30,
        description="Timeout in seconds until the request to the DSMS is timed out.",
    )

    username: Optional[SecretStr] = Field(
        None,
        description="User name for connecting to the DSMS instance",
    )
    password: Optional[SecretStr] = Field(
        None,
        description="Password for connecting to the DSMS instance",
    )
    token: Optional[SecretStr] = Field(
        None,
        description="JWT bearer token for connecting to the DSMS instance",
    )
    ssl_verify: bool = Field(
        True,
        description="Whether the SSL of the DSMS shall be verified during connection.",
    )

    encoding: str = Field(
        "utf-8",
        description="General encoding to be used for reading/writing serializations.",
    )

    datetime_format: str = Field(
        "%Y-%m-%dT%H:%M:%S.%f",
        description="Datetime format used in the DSMS instance.",
    )

    kitem_repo: str = Field(
        "knowledge",
        description="Repository of the triplestore for KItems in the DSMS",
    )

    @field_validator("token")
    def validate_auth(cls, val, info: ValidationInfo):
        """Validate the provided authentication/authorization secrets."""
        username = info.data.get("username")
        passwd = info.data.get("password")
        host_url = info.data.get("host_url")
        timeout = info.data.get("request_timeout")
        if username and passwd and val:
            raise ValueError(
                "Either `username` and `password` or `token` must be provided. Not both."
            )
        if username and not passwd:
            raise ValueError("`username` provided, but `password` not.")
        if not username and passwd:
            raise ValueError("`password` but not the `username` is defined.")
        if not username and not passwd and not val:
            warnings.warn(
                """No authentication details provided. Either `username` and `password`
                or `token` must be provided."""
            )
        if not val and username and passwd:
            url = urllib.parse.urljoin(str(host_url), "api/users/token")
            authorization = f"Basic {username.get_secret_value()}:{passwd.get_secret_value()}"
            response = requests.get(
                url,
                headers={"Authorization": authorization},
                timeout=timeout,
            )
            if not response.ok:
                raise RuntimeError(
                    f"Something went wrong fetching the access token: {response.text}"
                )
            val = response.json().get("token")
        if isinstance(val, str):
            if "Bearer " not in val:
                val = SecretStr(f"Bearer {val}")
            else:
                val = SecretStr(val)
        elif isinstance(val, SecretStr):
            if "Bearer " not in val.get_secret_value():
                val = SecretStr(f"Bearer {val.get_secret_value()}")

        return val

    model_config = SettingsConfigDict(env_prefix="DSMS_")



"""Most basic item in DSMS"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Dict, Optional
    from uuid import UUID

    from dsms.core.dsms import DSMS
    from dsms.knowledge.kitem import KItem


class Buffers:
    """Buffers of KItems for synchronization with the DSMS backend"""

    created: "Dict[UUID, KItem]" = {}

    updated: "Dict[UUID, KItem]" = {}

    deleted: "Dict[UUID, KItem]" = {}


class Context:
    """Object giving the current DSMS context."""

    dsms: "Optional[DSMS]" = None

    ktypes: "Dict[str, Any]" = {}

    buffers: Buffers = Buffers



"""App KProperty"""

from typing import TYPE_CHECKING, Optional

from pydantic import BaseModel, Field

from dsms.knowledge.properties.base import KProperty, KPropertyItem

if TYPE_CHECKING:
    from typing import Callable


class AdditionalProperties(BaseModel):
    """Additional properties of"""

    triggerUponUpload: bool = Field(
        False,
        description="Whether the app should be triggered when a file is uploaded",
    )
    triggerUponUploadFileExtensions: Optional[list[str]] = Field(
        None,
        description="File extensions for which the upload shall be triggered.",
    )


class App(KPropertyItem):
    """App of a KItem."""

    kitem_app_id: Optional[int] = Field(
        None, description="ID of the KItem App"
    )
    executable: str = Field(
        ..., description="Name of the executable related to the app"
    )
    title: Optional[str] = Field(None, description="Title of the appilcation")
    description: Optional[str] = Field(
        None, description="Description of the appilcation"
    )
    tags: Optional[dict] = Field(
        None, description="Tags related to the appilcation"
    )
    additional_properties: Optional[AdditionalProperties] = Field(
        None, description="Additional properties related to the appilcation"
    )

    def run(self, *args, **kwargs) -> None:
        """Run application"""
        raise NotImplementedError


class AppsProperty(KProperty):
    """KProperty for apps"""

    # OVERRIDE
    @property
    def k_property_item(cls) -> "Callable":
        """App data model"""
        return App

    # OVERRIDE
    def _add(self, item: App) -> App:
        """Side effect when an App is added to the KProperty"""
        return item

    # OVERRIDE
    def _update(self, item: App) -> App:
        """Side effect when an App is updated at the KProperty"""
        return item

    # OVERRIDE
    def _delete(self, item: App) -> None:
        """Side effect when deleting the App of a KItem"""

    # OVERRIDE
    def _get(self, item: App) -> App:
        """Side effect when getting the App for a specfic kitem"""
        return item



"""DSMS top module"""

from dsms.core.configuration import Configuration
from dsms.core.context import Context
from dsms.core.dsms import DSMS
from dsms.knowledge.kitem import KItem
from dsms.knowledge.ktype import KType

__all__ = ["DSMS", "Configuration", "Context", "KItem", "KType"]



"""DSMS Propertys module"""

from dsms.knowledge.properties.affiliations import (  # isort:skip
    Affiliation,
    AffiliationsProperty,
)
from dsms.knowledge.properties.annotations import (  # isort:skip
    Annotation,
    AnnotationsProperty,
)
from dsms.knowledge.properties.apps import App, AppsProperty
from dsms.knowledge.properties.authors import Author, AuthorsProperty
from dsms.knowledge.properties.base import KProperty, KPropertyItem
from dsms.knowledge.properties.contacts import ContactInfo, ContactsProperty
from dsms.knowledge.properties.custom_properties import CustomProperties
from dsms.knowledge.properties.hdf5 import Column, HDF5Container
from dsms.knowledge.properties.summary import Summary
from dsms.knowledge.properties.user_groups import UserGroup, UserGroupsProperty

from dsms.knowledge.properties.attachments import (  # isort:skip
    Attachment,
    AttachmentsProperty,
)


from dsms.knowledge.properties.linked_kitems import (  # isort:skip
    LinkedKItem,
    LinkedKItemsProperty,
)

from dsms.knowledge.properties.external_links import (  # isort:skip
    ExternalLink,
    ExternalLinksProperty,
)


__all__ = [
    "Annotation",
    "AnnotationsProperty",
    "Attachment",
    "AttachmentsProperty",
    "Author",
    "AuthorsProperty",
    "CustomProperties",
    "LinkedKItem",
    "LinkedKItemsProperty",
    "ContactInfo",
    "ContactsProperty",
    "ExternalLinksProperty",
    "ExternalLink",
    "AppsProperty",
    "App",
    "AffiliationsProperty",
    "Affiliation",
    "UserGroupsProperty",
    "UserGroup",
    "Summary",
    "KProperty",
    "KPropertyItem",
    "HDF5Container",
    "Column",
]



"""Custom properties of a KItem"""


from typing import TYPE_CHECKING, Any, Optional
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field, model_serializer

if TYPE_CHECKING:
    from typing import Set

    from dsms import Context


class Summary(BaseModel):
    """Model for the custom properties of the KItem"""

    id: Optional[UUID] = Field(None, description="ID of the KItem")
    text: str = Field(..., description="Summary text of the KItem")
    kitem: Optional[Any] = Field(
        None, description="KItem related to the summary", exclude=True
    )

    model_config = ConfigDict(
        extra="forbid", exclude={"kitem", "id"}, validate_assignment=True
    )

    def __setattr__(self, name, value) -> None:
        """Add kitem to updated-buffer if an attribute is set"""
        super().__setattr__(name, value)
        self._mark_as_updated()

    def __str__(self) -> str:
        """Pretty print the custom properties"""
        fields = ", ".join(
            [
                f"{key}={value}"
                for key, value in self.__dict__.items()
                if key not in self.exclude
            ]
        )
        return f"{self.__class__.__name__}({fields})"

    def __repr__(self) -> str:
        """Pretty print the custom properties"""
        return str(self)

    def __hash__(self) -> int:
        return hash(str(self))

    def _mark_as_updated(self) -> None:
        if self.kitem and self.id not in self.context.buffers.updated:
            self.context.buffers.updated.update({self.id: self.kitem})

    @property
    def id(cls) -> Optional[UUID]:
        """Identifier of the KItem related to the CustomProperies"""
        if not cls.kitem:
            raise ValueError("KItem not defined yet.")
        return cls.kitem.id  # pylint: disable=E1101

    @property
    def context(cls) -> "Context":
        """Getter for Context"""
        from dsms import (  # isort:skip
            Context,
        )

        return Context

    @property
    def exclude(cls) -> "Optional[Set[str]]":
        """Fields to be excluded from the JSON-schema"""
        return cls.model_config.get("exclude")

    @model_serializer
    def serialize(self) -> str:
        """Serialize the summary model"""
        return self.text



"""KItem types"""

from typing import Any, Dict, Optional, Union
from uuid import UUID

from pydantic import BaseModel, Field, field_validator, model_serializer

from dsms.knowledge.utils import _parse_model


class KType(BaseModel):
    """Knowledge type of the knowledge item."""

    id: Union[UUID, str] = Field(..., description="ID of the KType.")
    name: Optional[str] = Field(
        None, description="Human readable name of the KType."
    )
    form_data: Optional[Any] = Field(
        None, description="Form data of the KItem."
    )
    data_schema: Optional[Any] = Field(
        None, description="OpenAPI schema of the KItem."
    )

    @field_validator("data_schema")
    @classmethod
    def validate_data_schema(cls, value) -> Dict[str, Any]:
        """Validate the data schema of the ktype"""
        if isinstance(value, dict):
            value = _parse_model(value)
        elif not isinstance(value, BaseModel):
            raise TypeError(f"Invalid type for `data_schema`: {type(value)}")
        return value

    @model_serializer
    def serialize(self):
        """Serialize ktype."""
        return {
            key: (
                value
                if not isinstance(value, BaseModel)
                else value.model_dump_json()
            )
            for key, value in self.items()
        }



"""Basic property for a KItem"""
from abc import abstractmethod
from typing import TYPE_CHECKING, Optional
from uuid import UUID

from pydantic import (  # isort:skip
    BaseModel,
    ConfigDict,
    Field,
    PrivateAttr,
    model_serializer,
)

from dsms.core.utils import _snake_to_camel  # isort:skip
from dsms.knowledge.utils import _get_kitem  # isort:skip


if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Iterable, List, Set, Union

    from dsms import Context, KItem


class KPropertyItem(BaseModel):
    """List item for the KItem-property"""

    id: Optional[UUID] = Field(
        None, description="KItem ID related to the KPropertyItem"
    )

    model_config = ConfigDict(
        exclude={"id"},
        alias_generator=_snake_to_camel,
        populate_by_name=True,
        from_attributes=True,
    )

    _kitem = PrivateAttr(default=None)

    def __str__(self) -> str:
        """Pretty print the KProperty"""
        values = ", ".join(
            [
                f"{key}={value}"
                for key, value in self.__dict__.items()
                if key not in self.exclude
            ]
        )
        return f"{self.__class__.__name__}({values})"

    def __repr__(self) -> str:
        """Pretty print the KProperty"""
        return str(self)

    def __setattr__(self, index: int, item: "Any") -> None:
        """Add KItem to updated buffer."""
        if self._kitem and self.id not in self.context.buffers.updated:
            self.context.buffers.updated.update({self.id: self._kitem})
        super().__setattr__(index, item)

    def __hash__(self) -> int:
        return hash(str(self))

    @property
    def kitem(cls) -> "KItem":
        """KItem related to the KPropertyItem"""
        if not cls.id:
            raise ValueError("KItem not defined yet for KProperty")
        return _get_kitem(cls.id)

    @property
    def exclude(cls) -> "Optional[Set[str]]":
        """Fields to be excluded from the JSON-schema"""
        return cls.model_config.get("exclude")

    @model_serializer
    def serialize(self):
        """Serialize KPropertItem"""
        return {
            key: value for key, value in self.__dict__.items() if key != "id"
        }


class KProperty(list):
    """Basic class for an KItem-property."""

    def __init__(self, *args) -> None:
        self._kitem: "KItem" = None
        self.extend(args)

    @property
    @abstractmethod
    def k_property_item(cls) -> "Callable":
        """Return the KPropertyItem-class for the KProperty"""

    @abstractmethod
    def _add(self, item: KPropertyItem) -> KPropertyItem:
        """Side effect when an KPropertyItem is added to the KProperty"""

    @abstractmethod
    def _update(self, item: KPropertyItem) -> KPropertyItem:
        """Side effect when an KPropertyItem is updated in the KProperty"""

    @abstractmethod
    def _get(self, item: KPropertyItem) -> KPropertyItem:
        """Side effect when an KPropertyItem is retrieved from the KProperty"""

    @abstractmethod
    def _delete(self, item: KPropertyItem) -> None:
        """Side effect when an KPropertyItem is deleted from the KProperty"""

    def __str__(self) -> str:
        """Pretty print the KProperty"""
        values = ", \n".join(["\t\t" + repr(value) for value in self])
        if values:
            values = f"\n{values}\n\t"
        return f"[{values}]"

    def __repr__(self) -> str:
        """Pretty print the KProperty"""
        return str(self)

    def __hash__(self) -> int:
        return hash(str(self))

    def __setitem__(
        self, index: int, item: "Union[Dict, KPropertyItem]"
    ) -> None:
        """Add or Update KPropertyItem and add it to the updated-buffer."""

        self._mark_as_updated()
        item = self._check_k_property_item(item)
        if self.kitem:
            item.id = self.kitem.id
        try:
            if self[index] != item:
                item = self._update(item)
        except IndexError:
            item = self._add(item)
        super().__setitem__(index, item)

    def __delitem__(self, index: int) -> None:
        """Delete the KPropertyItem from the KProperty"""

        self._mark_as_updated()
        item = super().__delitem__(index)
        self._delete(item)

    def __getitem__(self, index: int) -> KPropertyItem:
        """Get the KPropertyItem from the KProperty"""

        item = super().__getitem__(index)
        return self._get(item)

    def __imul__(self, index: int) -> None:
        """Imul the KPropertyItem"""
        self._mark_as_updated()
        super().__imul__(index)

    def extend(self, iterable: "Iterable") -> None:
        """Extend KProperty with list of KPropertyItem"""
        from dsms import KItem

        to_extend = []
        for item in iterable:
            if isinstance(item, (list, tuple)):
                for subitem in item:
                    item = self._check_and_add_item(subitem)
                    to_extend.append(item)
            elif isinstance(item, (dict, KPropertyItem, KItem)):
                item = self._check_and_add_item(item)
                to_extend.append(item)
            else:
                to_extend.append(item)
        self._mark_as_updated()
        super().extend(to_extend)

    def append(self, item: "Union[Dict, Any]") -> None:
        """Append KPropertyItem to KProperty"""

        item = self._check_and_add_item(item)
        self._mark_as_updated()
        super().append(item)

    def insert(self, index: int, item: "Union[Dict, Any]") -> None:
        """Insert KPropertyItem at KProperty at certain index"""

        item = self._check_and_add_item(item)
        self._mark_as_updated()
        super().insert(index, item)

    def pop(self, index=-1) -> KPropertyItem:
        """Pop KPropertyItem from KProperty"""

        item = super().pop(index)
        self._mark_as_updated()
        self._delete(item)
        return item

    def remove(self, item: "Union[Dict, Any]") -> None:
        """Remove KPropertyItem from KProperty"""

        self._mark_as_updated()
        self._delete(item)
        super().remove(item)

    def _check_and_add_item(self, item: "Union[Dict, Any]") -> KPropertyItem:
        item = self._check_k_property_item(item)
        if self.kitem:
            item.id = self.kitem.id
        return self._add(item)

    def _check_k_property_item(
        self, item: "Union[Dict, Any]"
    ) -> KPropertyItem:
        """Check the type of the processsed KPropertyItem"""
        from dsms import KItem

        if not isinstance(item, (self.k_property_item, dict, KItem, str)):
            raise TypeError(
                f"""Item `{item}` must be of type {self.k_property_item}, {KItem}, {str} or {dict},
                not `{type(item)}`."""
            )
        if isinstance(item, dict):
            item = self.k_property_item(**item)
        return item

    def _mark_as_updated(self) -> None:
        """Add KItem of KProperty to updated buffer"""
        if self._kitem and self._kitem.id not in self.context.buffers.updated:
            self.context.buffers.updated.update({self._kitem.id: self._kitem})

    @property
    def context(cls) -> "Context":
        """Getter for Context"""
        from dsms import (  # isort:skip
            Context,
        )

        return Context

    @property
    def kitem(cls) -> "KItem":
        """KItem context of the field"""
        return cls._kitem

    @kitem.setter
    def kitem(cls, value: "KItem") -> None:
        """KItem setter"""
        cls._kitem = value
        for item in cls:
            item.id = cls.kitem.id

    @property
    def values(cls) -> "List[Dict[str, Any]]":
        """Values of the KProperty"""
        return list(cls)



"""KItem Apps utils"""

from typing import TYPE_CHECKING

from dsms.apps.apps import App
from dsms.core.utils import _perform_request

if TYPE_CHECKING:
    from typing import List


def _get_available_apps() -> "List[App]":
    """Get available KItem app."""
    response = _perform_request("api/knowledge/apps", "get")
    if not response.ok:
        message = f"""Something went wrong fetching the available app
        list in the DSMS: {response.text}"""
        raise RuntimeError(message)
    return [App(**app) for app in response.json()]



"""Author KProperty"""

from typing import TYPE_CHECKING
from uuid import UUID

from pydantic import Field

from dsms.knowledge.properties.base import KProperty, KPropertyItem

if TYPE_CHECKING:
    from typing import Callable


class Author(KPropertyItem):
    """Author of a KItem."""

    user_id: UUID = Field(..., description="ID of the DSMS User")


class AuthorsProperty(KProperty):
    """KProperty for authors"""

    # OVERRIDE
    @property
    def k_property_item(cls) -> "Callable":
        """Author data model"""
        return Author

    # OVERRIDE
    def _add(self, item: Author) -> Author:
        """Side effect when an Author is added to the KProperty"""
        return item

    # OVERRIDE
    def _update(self, item: Author) -> Author:
        """Side effect when an Author is updated at the KProperty"""
        return item

    # OVERRIDE
    def _get(self, item: Author) -> Author:
        """Side effect when getting the Author for a specfic kitem"""
        return item

    # OVERRIDE
    def _delete(self, item: Author) -> None:
        """Side effect when deleting the Author of a KItem"""
